#include "logreader.h"
#include "messages.h"
#include "fdread.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

static gboolean log_reader_fetch_log(LogReader *self, FDRead *fd);

typedef struct _LogReaderWatch
{
  GSource super;
  LogReader *reader;
  GPollFD pollfd;
  FDRead *fd;
} LogReaderWatch;

static gboolean
log_reader_fd_prepare(GSource *source,
                      gint *timeout)
{
  LogReaderWatch *self = (LogReaderWatch *) source;

  msg_debug("log_reader_fd_prepare(); window_size='%d'", self->reader->window_size);

  if (self->reader->mark_target != -1)
    {
      GTimeVal tv;
      
      g_source_get_current_time(source, &tv);
      *timeout = MAX((self->reader->mark_target - tv.tv_sec) * 1000, 0);
    }
  else
    {
      *timeout = -1;
    }
  self->pollfd.revents = 0;
  self->pollfd.events = 0;
  
  /* never indicate readability if flow control prevents us from sending messages */
  
  if (!self->reader->window_size)
    {
      return FALSE;
    }
  
  /* always readable if we have a complete line in our buffer.  as the
   * reader is at a lower priority than writers, buffers are flushed prior
   * to new messages are read
   */
  
  if (self->reader->flags & LR_COMPLETE_LINE)
    return TRUE;
  
  if (self->reader->flags & LR_FOLLOW)
    {
      *timeout = self->reader->options->follow_freq * 1000;
      return FALSE;
    }
  
  self->pollfd.events = G_IO_IN;
  return FALSE;
}

static gboolean
log_reader_fd_check(GSource *source)
{
  LogReaderWatch *self = (LogReaderWatch *) source;
  GTimeVal tv;

  g_source_get_current_time(source, &tv);
  
  if (self->reader->mark_target != -1 && self->reader->mark_target <= tv.tv_sec)
    return TRUE;

  if (self->reader->flags & LR_COMPLETE_LINE)
    return TRUE;
  
  if (self->reader->flags & LR_FOLLOW)
    {
      struct stat st;
      off_t pos;
      
      pos = lseek(self->fd->fd, 0, SEEK_CUR);
      if (pos == -1)
        return FALSE;
      
      if (fstat(self->fd->fd, &st) == -1)
	return FALSE;

      if (pos != st.st_size)
	return TRUE;
      return FALSE;
    }
    
  return !!(self->pollfd.revents & (G_IO_IN | G_IO_ERR | G_IO_HUP));
}

static gboolean
log_reader_fd_dispatch(GSource *source,
                       GSourceFunc callback,
                       gpointer user_data)
{
  LogReaderWatch *self = (LogReaderWatch *) source;
  GTimeVal tv;
  
  g_source_get_current_time(source, &tv);
  
  if (self->reader->mark_target != -1 && self->reader->mark_target <= tv.tv_sec)
    {
      log_pipe_queue(&self->reader->super, log_msg_new_mark(), PF_FLOW_CTL_OFF);
    }

  if (self->reader->options->mark_freq > 0)
    {
      g_source_get_current_time(source, &tv);
      self->reader->mark_target = tv.tv_sec + self->reader->options->mark_freq;
    }

  if (!log_reader_fetch_log(self->reader, self->fd))
    {
      return FALSE;
    }
    
  return TRUE;
}

static void
log_reader_fd_finalize(GSource *source)
{
  LogReaderWatch *self = (LogReaderWatch *) source;

  fd_read_free(self->fd);
  log_pipe_unref(&self->reader->super);
}

GSourceFuncs log_reader_source_funcs =
{
  log_reader_fd_prepare,
  log_reader_fd_check,
  log_reader_fd_dispatch,
  log_reader_fd_finalize
};

static GSource *
log_reader_watch_new(LogReader *reader, FDRead *fd)
{
  LogReaderWatch *self = (LogReaderWatch *) g_source_new(&log_reader_source_funcs, sizeof(LogReaderWatch));
  
  log_pipe_ref(&reader->super);
  self->reader = reader;
  self->fd = fd;
  self->pollfd.fd = fd->fd;
  if (self->reader->flags & LR_FOLLOW)
    {
    
      /* seek to eof, would be better to record persistently where we
       * stopped last time and seek there, in a way similar to what logcheck
       * does */
      
      lseek(fd->fd, 0, SEEK_END);
    }
  g_source_set_priority(&self->super, LOG_PRIORITY_READER);
  g_source_add_poll(&self->super, &self->pollfd);
  return &self->super;
}

static void
log_reader_msg_ack(LogMessage *lm, gpointer user_data)
{
  LogReader *self = (LogReader *) user_data;
  
  log_msg_ack_block_end(lm);
  self->window_size++;
  log_msg_unref(lm);
  
  /* as we are the source we don't ack the message */
}

static gboolean
log_reader_handle_line(LogReader *self, gchar *line, gint length, GSockAddr *saddr)
{
  LogMessage *m;
  guint parse_flags = 0;
  
  msg_debug("Incoming log entry; line='%.*s'", length, line);
  
  if (self->options->options & LRO_NOPARSE)
    parse_flags |= LP_NOPARSE;
  if (self->options->options & LRO_CHECK_HOSTNAME)
    parse_flags |= LP_CHECK_HOSTNAME;
  if (self->flags & LR_STRICT)
    parse_flags |= LP_STRICT;
    
  m = log_msg_new(line, length, saddr, parse_flags);
  
  if (self->flags & LR_INTERNAL)
    m->flags |= LF_INTERNAL;
  
  log_msg_ack_block_start(m, log_reader_msg_ack, self);
  self->window_size--;

  if (self->flags & LR_LOCAL)
    {
      m->flags |= LF_LOCAL;
    }    
  if (self->options->zone_offset_set)
    m->stamp.zone_offset = self->options->zone_offset;
  log_pipe_queue(self->super.pipe_next, m, 0);
  
  return !!self->window_size;
}

static gboolean
log_reader_iterate_buf(LogReader *self, GSockAddr *saddr)
{
  gchar *eol, *start;
  gint length, msg_count;
  gboolean may_read = TRUE;

  self->flags &= ~LR_COMPLETE_LINE;
  eol = memchr(self->buffer, '\0', self->ofs);
  if (eol == NULL)
    eol = memchr(self->buffer, '\n', self->ofs);

  if (!eol &&
      ((self->ofs == self->options->msg_size) || 
       ((self->flags & LR_PKTTERM) && self->ofs) || 
       self->options->padding)) 
    {
      /* our buffer is full, or
       * we are set to packet terminating mode and there's no terminating new line, or
       * we are in padded mode HP-UX
       */
      log_reader_handle_line(self, self->buffer, 
			     (self->options->padding 
			      ? (eol ? eol - self->buffer : self->ofs)
			      : self->ofs), 
			     saddr);
      self->ofs = 0;
    }
  else
    {
      if (saddr == NULL)
        {
          saddr = self->prev_addr;
          self->prev_addr = NULL;
        }
      msg_count = 0;
      start = self->buffer;
      while ((self->options->fetch_limit == 0 || msg_count < self->options->fetch_limit) && eol && may_read)
	{
	  gchar *end = eol;
	  /* eol points at the newline character. end points at the
	   * character terminating the line, which may be a carriage
	   * return preceeding the newline. */
	   
	  msg_count++;
	  
	  while ((end > start) && (end[-1] == '\r' || end[-1] == '\n' || end[-1] == 0))
	    end--;
	  
	  length = end - start;
	  
	  if (length)
	    may_read = log_reader_handle_line(self, start, length, saddr);
	  
	  start = eol + 1;

	  eol = memchr(start, '\0', &self->buffer[self->ofs] - start);
	  if (eol == NULL)
	    eol = memchr(start, '\n', &self->buffer[self->ofs] - start);
	}
      
      /* move remaining data to the beginning of the buffer */
      
      if (eol)
        self->flags |= LR_COMPLETE_LINE;
      
      self->ofs = &self->buffer[self->ofs] - start;
      memmove(self->buffer, start, self->ofs);
      if (self->ofs != 0)
        self->prev_addr = saddr;
      else
        g_sockaddr_unref(saddr);
    }

  return TRUE;
}

static gboolean
log_reader_fetch_log(LogReader *self, FDRead *fd)
{
  guint avail = self->options->msg_size - self->ofs;
  gint rc;
  GSockAddr *sa = NULL;

  /* iterare on previously buffered data */
  if (self->flags & LR_COMPLETE_LINE)
    {
      log_reader_iterate_buf(self, NULL);
      
      /* we still have something */
      if (self->ofs != 0)
        return TRUE;
    }
  
  /* check for new data */
  if (self->options->padding)
    {
      if (avail < self->options->padding)
	{
	  msg_error("Buffer is too small to hold padding number of bytes; fd='%d', padding='%d', avail='%d'", fd->fd, self->options->padding, avail);
          log_pipe_notify(self->control, &self->super, NC_CLOSE, self);
	  return FALSE;
	}
      avail = self->options->padding;
    }
  rc = fd_read(fd, self->buffer + self->ofs, avail, &sa);

  if (rc == -1)
    {
      /* an error occurred while reading */
      msg_error("I/O error occurred while reading; fd='%d', error='%m'", fd->fd);
      log_pipe_notify(self->control, &self->super, NC_READ_ERROR, self);
      g_sockaddr_unref(sa);
      return FALSE;
    }
  else if (rc == 0 && (self->flags & LR_FOLLOW) == 0)
    {
      /* EOF read */
      msg_verbose("EOF occurred while reading; fd='%d'", fd->fd);
      log_pipe_notify(self->control, &self->super, NC_CLOSE, self);
      g_sockaddr_unref(sa);
      return FALSE;
    }
  else 
    {
      if (self->options->padding && rc != self->options->padding)
	{
	  msg_error("Padding was set, and couldn't read enough bytes; fd='%d', padding='%d', read='%d'", fd->fd, self->options->padding, rc);
          log_pipe_notify(self->control, &self->super, NC_READ_ERROR, self);
          g_sockaddr_unref(sa);
	  return FALSE;
	}
      self->ofs += rc;
      log_reader_iterate_buf(self, sa);
    }

  return TRUE;
}

static gboolean
log_reader_init(LogPipe *s, GlobalConfig *cfg, PersistentConfig *persist)
{
  LogReader *self = (LogReader *) s;
  
  /* the source added below references this logreader, it will be unref'd
     when the source is destroyed */ 
  self->source = log_reader_watch_new(self, self->fd);
  g_source_attach(self->source, NULL);
  return TRUE;
}

static gboolean
log_reader_deinit(LogPipe *s, GlobalConfig *cfg, PersistentConfig *persist)
{
  LogReader *self = (LogReader *) s;
  
  if (self->source)
    {
      g_source_destroy(self->source);
      g_source_unref(self->source);
      self->source = NULL;
    }
  return TRUE;
}

static void
log_reader_free(LogPipe *s)
{
  LogReader *self = (LogReader *) s;
  
  /* when this function is called the source is already removed, because it
     holds a reference to this reader */
  log_pipe_unref(self->control);
  g_free(self->buffer);
  g_free(s);
}

LogPipe *
log_reader_new(FDRead *fd, guint32 flags, LogPipe *control, LogReaderOptions *options)
{
  LogReader *self = g_new0(LogReader, 1);

  log_pipe_init_instance(&self->super);
  self->super.init = log_reader_init;
  self->super.deinit = log_reader_deinit;
  self->super.free_fn = log_reader_free;
  self->options = options;
  self->flags = flags;
  self->window_size = options->init_window_size;
  self->fd = fd;
  log_pipe_ref(control);
  self->control = control;
  self->buffer = g_malloc(options->msg_size);
  self->mark_target = -1;
  if (options->follow_freq > 0)
    self->flags |= LR_FOLLOW;
  return &self->super;
}

void
log_reader_options_defaults(LogReaderOptions *options)
{
  options->padding = 0;
  options->init_window_size = -1;
  options->options = 0;
  options->fetch_limit = -1;
  options->msg_size = -1;
  options->follow_freq = -1; 
  options->mark_freq = -1;
  options->zone_offset = timezone;
}

void
log_reader_options_init(LogReaderOptions *options, GlobalConfig *cfg)
{
  if (options->fetch_limit == -1)
    options->fetch_limit = cfg->log_fetch_limit;
  if (options->init_window_size == -1)
    options->init_window_size = cfg->log_iw_size;
  if (options->msg_size == -1)
    options->msg_size = cfg->log_msg_size;
  if (options->mark_freq == -1)
    options->mark_freq = cfg->mark_freq;
  if (options->follow_freq == -1)
    options->follow_freq = cfg->follow_freq;
}

