#include "macros.h"
#include "logmsg.h"
#include "syslog-names.h"
#include "messages.h"

#include "filter.h"

#include <time.h>

#define M_FACILITY 1
#define M_LEVEL    2
#define M_TAG      3

#define M_DATE     4
#define M_FULLDATE 5
#define M_ISODATE  6
#define M_YEAR     7
#define M_MONTH    8
#define M_DAY      9
#define M_HOUR     10
#define M_MIN      11
#define M_SEC      12
#define M_WEEKDAY  13

#define M_FULLHOST 14
#define M_HOST     15
#define M_PROGRAM  16

#define M_MESSAGE  17

struct macro_def
{
  char *name;
  int id;
}
macros[] =
{
  { "FACILITY", M_FACILITY },
  { "PRIORITY", M_LEVEL },
  { "LEVEL", M_LEVEL },
  { "TAG", M_TAG },
  { "DATE", M_DATE },
  { "FULLDATE", M_FULLDATE },
  { "ISODATE", M_ISODATE },
  { "YEAR", M_YEAR },
  { "MONTH", M_MONTH },
  { "DAY", M_DAY },
  { "HOUR", M_HOUR },
  { "MIN", M_MIN },
  { "SEC", M_SEC },
  { "WEEKDAY", M_WEEKDAY },
  { "HOST", M_HOST },
  { "FULLHOST", M_FULLHOST },
  { "PROGRAM", M_PROGRAM },
  { "MSG", M_MESSAGE },
  { "MESSAGE", M_MESSAGE }
};

GHashTable *macro_hash;

static void
result_append(GString *result, gchar *str, gint len, gboolean escape)
{
  gint i;
  
  if (escape)
    {
      for (i = 0; i < len; i++)
        {
          if (str[i] == '\'' || str[i] == '"' || str[i] == '\\')
            {
              g_string_append_c(result, '\\');
              g_string_append_c(result, str[i]);
            }
          else if (str[i] < ' ')
            {
              g_string_sprintfa(result, "\\%03o", (unsigned int) str[i]);
            }
          else
            g_string_append_c(result, str[i]);
        }
    }
  else
    g_string_append_len(result, str, len);
    
}

static gboolean
expand_macro(GString *result, gint id, guint32 flags, glong zone_offset, LogMessage *msg)
{
  switch (id)
    {
    case M_FACILITY:
      {
        /* facility */
        char *n;
        
        n = syslog_lookup_value(msg->pri & LOG_FACMASK, sl_facilities);
        if (n)
          {
            g_string_append(result, n);
          }
        else
          {
            g_string_sprintfa(result, "%x", (msg->pri & LOG_FACMASK) >> 3);
          }
        break;
      }
    case M_LEVEL:
      {
        /* level */
        char *n;
        
        n = syslog_lookup_value(msg->pri & LOG_PRIMASK, sl_levels);
        if (n)
          {
            g_string_append(result, n);
          }
        else
          {
            g_string_sprintfa(result, "%d", msg->pri & LOG_PRIMASK);
          }

        break;
      }
    case M_TAG:
      {
        g_string_sprintfa(result, "%02x", msg->pri);
        break;
      }
    case M_FULLHOST:
      {
        /* full hostname */
        result_append(result, msg->host->str, msg->host->len, !!(flags & MF_ESCAPE_RESULT));
        break;
      }
    case M_HOST:
      {
        /* host */
        gchar *p1 = memchr(msg->host->str, '@', msg->host->len);
        gchar *p2;
        int remaining, length;

        if (p1)
          p1++;
        else
          p1 = msg->host->str;
        remaining = msg->host->len - (p1 - msg->host->str);
        p2 = memchr(p1, '/', remaining);
        length = p2 ? p2 - p1 
                    : msg->host->len - (p1 - msg->host->str);
        result_append(result, p1, length, !!(flags & MF_ESCAPE_RESULT));
        break;
      }
    case M_PROGRAM:
      {
        /* program */
        if (msg->program->len)
          {
            result_append(result, msg->program->str, msg->program->len, !!(flags & MF_ESCAPE_RESULT));
          }
        break;
      }
    case M_FULLDATE:
    case M_ISODATE:
    case M_WEEKDAY:
    case M_DATE:
    case M_YEAR:
    case M_MONTH:
    case M_DAY:
    case M_HOUR:
    case M_MIN:
    case M_SEC:
      {
        /* year, month, day */
        struct tm *tm;
        gchar buf[64];
        gint length;

        if (flags & MF_STAMP_RECVD)
          tm = gmtime(&msg->recvd.time.tv_sec - zone_offset);
        else
          tm = gmtime(&msg->stamp.time.tv_sec - zone_offset);

        switch (id)
          {
          case M_WEEKDAY:
            length = strftime(buf, sizeof(buf), "%a", tm);
            g_string_append_len(result, buf, length);
            break;
          case M_YEAR:
            g_string_sprintfa(result, "%04d", tm->tm_year + 1900);
            break;
          case M_MONTH:
            g_string_sprintfa(result, "%02d", tm->tm_mon + 1);
            break;
          case M_DAY:
            g_string_sprintfa(result, "%02d", tm->tm_mday);
            break;
          case M_HOUR:
            g_string_sprintfa(result, "%02d", tm->tm_hour);
            break;
          case M_MIN:
            g_string_sprintfa(result, "%02d", tm->tm_min);
            break;
          case M_SEC:
            g_string_sprintfa(result, "%02d", tm->tm_sec);
            break;
          case M_ISODATE:
            length = strftime(buf, sizeof(buf), "%Y-%m-%dT%H:%M:%S%z", tm);
            g_string_append_len(result, buf, length);
            break;
          case M_FULLDATE:
            length = strftime(buf, sizeof(buf), "%Y %h %e %H:%M:%S", tm);
            g_string_append_len(result, buf, length);
            break;
          case M_DATE:
            g_string_append_len(result, msg->date->str, msg->date->len);
            break;
          }
        break;
      }
    case M_MESSAGE:
      /* message */
      result_append(result, msg->msg->str, msg->msg->len, !!(flags & MF_ESCAPE_RESULT));
      break;
    default:
      msg_fatal("Internal error, unknown macro referenced...");
      return FALSE;
    }
  return TRUE;
}

GString *
expand_macros(GString *result, gchar *template, guint32 flags, glong zone_offset, LogMessage *msg)
{
  gchar *p;

  if (!macro_hash)
    {
      int i;
      macro_hash = g_hash_table_new(g_str_hash, g_str_equal);
      for (i = 0; i < sizeof(macros) / sizeof(macros[0]); i++)
        {
          g_hash_table_insert(macro_hash, macros[i].name,
                              (gpointer) macros[i].id);
        }
    }

  g_string_truncate(result, 0);
  p = template;
  while (*p)
    {
      if (*p == '$')
        {
          /* macro begins */
          gchar macro[32];
          int i = 0;

          p++;
          if (*p >= '0' && *p <= '9')
            {
              gint ndx = *p - '0';
              /* match reference */
              if (re_matches[ndx])
                result_append(result, re_matches[ndx], strlen(re_matches[ndx]), !!(flags & MF_ESCAPE_RESULT));
              p++;
            }
          else
            {
              while ((i < sizeof(macro) - 1) &&
                     ((*p >= 'A' && *p <= 'Z') || (*p == '_')))
                {
                  macro[i] = *p;
                  i++;
                  p++;
                }
              macro[i] = 0;

              i = (gint) g_hash_table_lookup(macro_hash, macro);
              if (i != 0)
                {
                  expand_macro(result, i, flags, zone_offset, msg);
                }
            }
        }
      else
        {
          /* otherwise simply copy */
          g_string_append_c(result, *p);
          p++;
        }
    }
  return result;
}
